use std::time;

use async_trait::async_trait;
use chrono::Utc;
use log::warn;
use uuid::Uuid;

use crate::{
    message_handler::user_commands::{
        movement::attempt_move_immediate, CommandHandlingError, UResult,
    },
    models::{
        item::{Item, SkillType},
        task::{Task, TaskDetails, TaskMeta},
    },
    regular_tasks::{
        queued_command::{MovementSource, QueueCommand, QueuedCommandContext},
        TaskHandler, TaskRunContext,
    },
    static_content::room::{room_map_by_code, Direction, MaterialType, Room},
    DResult,
};

use super::{combat::change_health, comms::broadcast_to_room, skills::skill_check_and_grind};

pub struct EnvironmentHandler;

pub async fn ensure_appropriate_environment_handler_after_movement(
    ctx: &mut QueuedCommandContext<'_>,
) -> UResult<()> {
    if ctx.item.item_type != "player" {
        return Ok(());
    }
    let (loc_type, loc_code) = match ctx.item.location.split_once("/") {
        None => return Ok(()),
        Some(v) => v,
    };
    let need_handler = if loc_type != "room" {
        false
    } else {
        let room = match room_map_by_code().get(loc_code) {
            None => return Ok(()),
            Some(r) => r,
        };
        match room.material_type {
            MaterialType::WaterSurface | MaterialType::Underwater => true,
            _ => false,
        }
    };

    let existing_task = ctx
        .trans
        .fetch_specific_task("ApplyEnvironmentalEffects", &ctx.item.refstr())
        .await?;
    if need_handler {
        if existing_task.is_none() {
            ctx.trans
                .upsert_task(&Task {
                    meta: TaskMeta {
                        task_code: ctx.item.refstr(),
                        next_scheduled: Utc::now() + chrono::TimeDelta::try_seconds(10).unwrap(),
                        ..Default::default()
                    },
                    details: TaskDetails::ApplyEnvironmentalEffects {
                        on_player_type: ctx.item.item_type.clone(),
                        on_player_code: ctx.item.item_code.clone(),
                    },
                })
                .await?;
        }
    } else {
        if existing_task.is_some() {
            ctx.trans
                .delete_task("ApplyEnvironmentalEffects", &ctx.item.refstr())
                .await?;
        }
    }
    Ok(())
}

#[async_trait]
impl TaskHandler for EnvironmentHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        let (player_type, player_code) = match &ctx.task.details {
            TaskDetails::ApplyEnvironmentalEffects {
                on_player_type,
                on_player_code,
            } => (on_player_type, on_player_code),
            x => {
                warn!(
                    "EnvironmentHandler called with bad TaskDetail type: {:#?}",
                    x
                );
                return Ok(None);
            }
        };

        let player = match ctx
            .trans
            .find_item_by_type_code(&player_type, &player_code)
            .await?
        {
            None => return Ok(None),
            Some(p) => p,
        };
        let mut player = (*player).clone();

        let (loc_type, loc_code) = match player.location.split_once("/") {
            None => return Ok(None),
            Some(v) => v,
        };
        if loc_type != "room" {
            return Ok(None);
        };

        let room = match room_map_by_code().get(loc_code) {
            None => return Ok(None),
            Some(r) => r,
        };

        let need_save = water_environment_effects(ctx, &room, &mut player).await?;

        if need_save {
            ctx.trans.save_item_model(&player).await?;
        }

        Ok(Some(time::Duration::from_secs(10)))
    }
}

pub async fn water_environment_effects(
    ctx: &mut TaskRunContext<'_>,
    room: &Room,
    player_item: &mut Item,
) -> DResult<bool> {
    match room.material_type {
        MaterialType::WaterSurface => {}
        MaterialType::Underwater => {
            if change_health(
                &ctx.trans,
                -5,
                player_item,
                &format!(
                    "{} can't breathe underwater",
                    player_item.display_for_sentence(1, true)
                ),
            )
            .await?
            {
                return Ok(true);
            }
        }
        _ => return Ok(false),
    }
    match room.exits.iter().find(|v| v.direction == Direction::DOWN) {
        None => return Ok(false),
        Some(_) => {
            if player_item.active_climb.is_some() {
                return Ok(false);
            }
            if skill_check_and_grind(&ctx.trans, player_item, &SkillType::Swim, 9.0).await? > 0.0 {
                return Ok(true);
            } else {
                broadcast_to_room(
                    &ctx.trans,
                    &player_item.location,
                    None,
                    &format!(
                        "{} seems to be sinking.\n",
                        player_item.display_for_sentence(1, true)
                    ),
                )
                .await?;
                player_item.queue.truncate(0);
                let source = MovementSource::Internal {
                    event_id: Uuid::new_v4(),
                };
                let mut qctx = QueuedCommandContext {
                    trans: ctx.trans,
                    command: &QueueCommand::Movement {
                        direction: Direction::DOWN,
                        source: source.clone(),
                    },
                    item: player_item,
                };
                match attempt_move_immediate(&Direction::DOWN, &mut qctx, &source).await {
                    Ok(_) => {}
                    Err(CommandHandlingError::UserError(_)) => {}
                    Err(CommandHandlingError::SystemError(e)) => return Err(e),
                }
            }
        }
    }
    Ok(true)
}

pub static TASK_HANDLER: &'static (dyn TaskHandler + Sync + Send) = &EnvironmentHandler;
